from __future__ import annotations
import asyncio
import dataclasses
from decimal import Decimal
import itertools
import pprint
import typing
from types import TracebackType
from typing import AsyncGenerator, Iterable, Optional, Sequence

import cachebox
from cachebox import Cache
from eth_typing import ChecksumAddress
from mach_client import (
    AccountManager,
    AssetServer,
    ChainClient,
    MachClient,
    RiskManager,
    Token,
)
from mach_client.account import EthereumAccount, EthereumAccountID
from mach_client.asset import EthereumToken
from mach_client.chain import EthereumChain
from mach_client.chain_client import EthereumClient
from mach_client.client.event import Trade
from mach_client.transaction import EthereumTransaction
from web3 import Web3
from web3.contract import AsyncContract

from .event import (
    AaveEvent,
    ConvertError,
    LiquidityRateError,
    RebalanceEvaluation,
    Supply,
    SupplyError,
    Withdraw,
    WithdrawError,
)
from .rebalance_manager import RebalanceManager
from .. import config, mach
from ..log import LogContextAdapter, Logger
from ..mach.destination_policy import TokenIteratorPolicy


LIQUIDITY_RATE_SCALING_FACTOR = 10**27


@dataclasses.dataclass(kw_only=True, slots=True)
class Aave:
    client: MachClient
    asset_server: AssetServer
    logger: Logger
    # Mapping of token to aToken
    atokens: dict[EthereumToken, EthereumToken] = {}
    # The Ethereum tokens (not atokens) that the Aave strategy will actually use
    enabled_tokens: list[EthereumToken] = []

    async def _update_chain(self, chain: EthereumChain) -> None:
        client = typing.cast(EthereumClient, await ChainClient.create(chain))
        await self.update_atokens(client)

    @classmethod
    async def create(
        cls,
        client: MachClient,
        asset_server: AssetServer,
        chains: Iterable[EthereumChain],
        logger: Logger,
    ) -> Aave:
        logger = LogContextAdapter(logger, "Aave")

        self = cls(client=client, asset_server=asset_server, logger=logger)

        self.logger.info("Fetching Aave data")
        await asyncio.gather(*[self._update_chain(chain) for chain in chains])
        self.logger.info("Done fetching Aave data")

        self.enabled_tokens = [
            token
            for token in self.valid_tokens
            if token.symbol in config.config.aave.symbols
        ]

        self.logger.info("Tokens:")
        self.logger.info(pprint.pformat(self.enabled_tokens))

        return self

    # The tokens available for use in Aave pools
    @property
    def valid_tokens(self) -> Sequence[EthereumToken]:
        return tuple(self.atokens.keys())

    def pool_addresses_provider(self, client: EthereumClient) -> AsyncContract:
        address = config.config.aave.pool_addresses_provider[client.chain]
        assert Web3.is_checksum_address(address)

        return client.w3.eth.contract(
            address=typing.cast(ChecksumAddress, address),
            abi=config.aave_pool_addresses_provider_abi,
        )

    async def protocol_data_provider(self, client: EthereumClient) -> AsyncContract:
        pool_addresses_provider = self.pool_addresses_provider(client)

        address = await pool_addresses_provider.functions.getPoolDataProvider().call()

        return client.w3.eth.contract(
            address=address,
            abi=config.aave_protocol_data_provider_abi,
        )

    async def pool(self, client: EthereumClient) -> AsyncContract:
        pool_addresses_provider = self.pool_addresses_provider(client)

        address = await pool_addresses_provider.functions.getPool().call()

        return client.w3.eth.contract(
            address=address,
            abi=config.aave_pool_abi,
        )

    @cachebox.cached(Cache(0))
    async def update_atokens(self, client: EthereumClient) -> None:
        protocol_data_provider = await self.protocol_data_provider(client)

        # Tuples (symbol, address) where the symbol is of the form "a<first 3 letters of chain name><symbol name>", ie. aArbUSDT
        raw_atokens: list[
            tuple[str, ChecksumAddress]
        ] = await protocol_data_provider.functions.getAllATokens().call()

        atokens = await asyncio.gather(
            *[
                EthereumToken.from_data(client, address, symbol, None)
                for symbol, address in raw_atokens
            ]
        )

        for atoken in atokens:
            # Aave has this weird thing with USDC where the atoken with symbol "USDC" actually represents a wrapped version
            # The "USDCn" token represents your actual USDC balance in the pool, where "n" stands for "native"
            symbol = "USDC" if atoken.symbol.endswith("USDCn") else atoken.symbol[4:]

            if not (
                token := typing.cast(
                    EthereumToken, Token.try_lookup_symbol(client.chain, symbol)
                )
            ):
                self.logger.warning(f"Could not find token {client.chain}-{symbol}")
                continue

            self.atokens[token] = atoken

    async def get_atoken_balance(
        self, token: EthereumToken, account_id: EthereumAccountID
    ) -> int:
        if token not in self.atokens:
            await self.update_atokens(token.client)

        atoken = self.atokens[token]

        return await atoken.get_balance(account_id)

    async def get_atoken_balance_in_coins(
        self, token: EthereumToken, account_id: EthereumAccountID
    ) -> Decimal:
        balance = await self.get_atoken_balance(token, account_id)
        return self.atokens[token].to_coins(balance)

    async def get_liquidity_rate(self, token: EthereumToken) -> Decimal:
        pool = await self.pool(token.client)
        reserve_data = await pool.functions.getReserveData(token.address).call()
        return Decimal(reserve_data[2]) / LIQUIDITY_RATE_SCALING_FACTOR

    # Returned dict is ordered by liquidity rate, descending
    async def get_liquidity_rates(
        self,
        tokens: Sequence[EthereumToken],
    ) -> dict[EthereumToken, Decimal]:
        liquidity_rates = await asyncio.gather(*map(self.get_liquidity_rate, tokens))

        result = list(zip(tokens, liquidity_rates))
        result.sort(key=lambda x: x[1], reverse=True)

        return dict(result)

    async def supply(
        self, token: EthereumToken, account: EthereumAccount
    ) -> tuple[int, Optional[Exception]]:
        logger = LogContextAdapter(self.logger, f"{token} => Aave")

        if (balance := await token.get_balance(account.downcast())) <= 0:
            logger.warning("Balance was empty, not supplying")
            return 0, None

        try:
            pool = await self.pool(token.client)

            supply_function = pool.functions.supply(
                token.address,
                balance,
                account.address,
                0,  # Referral code
            )

            # TODO: This should be `balance`, but that causes occasional errors like this:
            # web3.exceptions.ContractLogicError: ('execution reverted: ERC20: transfer amount exceeds allowance', '0x08c379a00000000000000000000000000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000000000000002845524332303a207472616e7366657220616d6f756e74206578636565647320616c6c6f77616e6365000000000000000000000000000000000000000000000000')
            # Maybe you can't withdraw any additional interest accrued within the current block even if it is counted in balanceOf?
            approval_amount = config.solidity_uint_max
            spender = EthereumAccountID(chain=token.chain, address_=pool.address)

            logger.info(f"Approving {approval_amount} units")

            await token.approve(
                account,
                spender,
                approval_amount,
            )

            logger.info(f"Supplying {balance} units")
            transaction = await EthereumTransaction.from_contract_function(
                token.client,
                supply_function,
                account,
            )
            sent_transaction = await transaction.broadcast()
            logger.debug("Transaction sent")

            receipt = await sent_transaction.wait_for_receipt()
            logger.debug("Receipt:")
            logger.debug(pprint.pformat(receipt))

        except Exception as e:
            return balance, e

        logger.info("Supply successful")

        return balance, None

    async def withdraw(
        self, token: EthereumToken, account: EthereumAccount
    ) -> tuple[int, Optional[Exception]]:
        logger = LogContextAdapter(self.logger, f"Aave {token} => account")

        if (balance := await self.get_atoken_balance(token, account.downcast())) <= 0:
            logger.debug("Balance was empty, not withdrawing")
            return 0, None

        try:
            pool = await self.pool(token.client)

            withdraw_function = pool.functions.withdraw(
                token.address,
                config.solidity_uint_max,  # Means withdraw everything
                account.address,
            )

            logger.info(f"Withdrawing {balance} units")
            transaction = await EthereumTransaction.from_contract_function(
                token.client,
                withdraw_function,
                account,
            )
            sent_transaction = await transaction.broadcast()
            logger.debug("Transaction sent")

            receipt = await sent_transaction.wait_for_receipt()
            logger.debug("Receipt:")
            logger.debug(pprint.pformat(receipt))

        except Exception as e:
            return balance, e

        logger.info("Withdraw successful")

        return balance, None

    async def run(
        self,
        *,
        accounts: AccountManager,
        rebalance_manager: RebalanceManager,
        # When rebalancing, if swapping to higher rate tokens fails, should we try swapping to a lower rate token or keep what we have?
        filter_lower_rate_tokens: bool,
        risk_manager: RiskManager,
        logger: Logger,
    ) -> AsyncGenerator[AaveEvent, None]:
        while True:
            # Inner loop determines when to rebalance portfolio
            while True:
                try:
                    token_rates = await self.get_liquidity_rates(self.enabled_tokens)
                except Exception as e:
                    logger.error(
                        "An exception was thrown while fetching liquidity rates from Aave:",
                        exc_info=e,
                    )
                    yield LiquidityRateError(self.enabled_tokens, e)
                    continue

                logger.info("Liquidity rates:")
                logger.info(pprint.pformat(token_rates))

                portfolio_balances = await asyncio.gather(
                    *[
                        self.get_atoken_balance_in_coins(
                            token,
                            typing.cast(
                                EthereumAccount, accounts[token.chain]
                            ).downcast(),
                        )
                        for token in self.enabled_tokens
                    ]
                )

                portfolio_balance_pairs = list(
                    zip(self.enabled_tokens, portfolio_balances)
                )

                logger.info("Portfolio balances:")
                logger.info(pprint.pformat(portfolio_balance_pairs))

                rebalance_analysis = rebalance_manager(
                    token_rates, portfolio_balance_pairs
                )

                yield RebalanceEvaluation(rebalance_analysis)

                if rebalance_analysis.rebalance:
                    break

                logger.info("Not rebalancing portfolio")
                await asyncio.sleep(config.config.aave.supply_duration)

            logger.info("Rebalancing portfolio")

            logger.info("Withdrawing funds from Aave")

            withdrawn = []

            for token in self.enabled_tokens:
                account = typing.cast(EthereumAccount, accounts[token.chain])
                amount, exception = await self.withdraw(token, account)

                if exception:
                    logger.error(
                        f"An exception was thrown while withdrawing {token} from Aave:",
                        exc_info=exception,
                    )
                    yield WithdrawError(token, token.to_coins(amount), exception)
                    continue

                elif amount <= 0:
                    continue

                withdrawn.append((token, token.to_coins(amount)))

            yield Withdraw(withdrawn)

            logger.info("Swapping funds in wallet")

            for src_token, rate in token_rates.items():
                account = typing.cast(EthereumAccount, accounts[src_token.chain])

                if await src_token.get_balance(account.downcast()) <= 0:
                    continue

                if filter_lower_rate_tokens:
                    next_tokens = itertools.takewhile(
                        lambda item: item[1] > rate, token_rates.items()
                    )
                else:
                    next_tokens = token_rates.items()

                if not next_tokens:
                    continue

                destination_policy = TokenIteratorPolicy(
                    self.asset_server,
                    map(lambda item: item[0], next_tokens),
                )

                runner = mach.run(
                    client=self.client,
                    asset_server=self.asset_server,
                    src_token=src_token,
                    destination_policy=destination_policy,
                    accounts=accounts,
                    risk_manager=risk_manager,
                    logger=logger,
                )

                try:
                    async for event in runner:
                        if isinstance(event, Trade):
                            break

                        logger.error(
                            f"Unexpected event while swapping out of {src_token}:"
                        )
                        logger.error(pprint.pformat(event))

                        yield ConvertError(src_token, event)

                except Exception as e:
                    logger.error(
                        f"An exception was thrown while swapping {src_token}:",
                        exc_info=e,
                    )
                    yield ConvertError(src_token, e)

            supplied = []

            for token in self.enabled_tokens:
                account = typing.cast(EthereumAccount, accounts[token.chain])
                amount, exception = await self.supply(token, account)

                if exception:
                    logger.error(
                        f"An exception was thrown while supplying {token} to Aave:",
                        exc_info=exception,
                    )
                    yield SupplyError(
                        token, Decimal(amount) / 10**token.decimals, exception
                    )
                    continue
                elif amount <= 0:
                    continue

                supplied.append((token, Decimal(amount) / 10**token.decimals))

            yield Supply(supplied)

            if not supplied:
                logger.warning("No tokens were supplied. Trying again.")
                continue

            logger.info("Sleeping...")
            await asyncio.sleep(config.config.aave.supply_duration)


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class AaveRunner:
    aave: Aave
    accounts: AccountManager
    rebalance_manager: RebalanceManager
    filter_lower_rate_tokens: bool
    risk_manager: RiskManager
    logger: Logger

    async def __aenter__(self) -> AsyncGenerator[AaveEvent, None]:
        return self.aave.run(
            accounts=self.accounts,
            rebalance_manager=self.rebalance_manager,
            filter_lower_rate_tokens=self.filter_lower_rate_tokens,
            risk_manager=self.risk_manager,
            logger=self.logger,
        )

    # Withdraw all tokens when the runner is done
    async def __aexit__(
        self,
        _exc_type: Optional[type[BaseException]],
        _exc_value: Optional[BaseException],
        _traceback: Optional[TracebackType],
    ) -> None:
        for token in self.aave.enabled_tokens:
            account = typing.cast(EthereumAccount, self.accounts[token.chain])
            _amount, exception = await self.aave.withdraw(token, account)

            if exception:
                self.logger.error(
                    f"Withdrawal of {token.symbol} from Aave failed with exception:",
                    exc_info=exception,
                )
